<?php


/**
 * SONIA CACHE
 * Use this cache by making any of your classes extend this class.
 * For use with Drupal, but could easily be adapted to an external application.
 * 
 * NOTES:
 * use for strings under 2MB
 * 
 * FEATURES
 * 1. The most important feature of this caching is that the keys can contain
 * substrings. The key can be created using an array, then that array is used to create
 * an ordered key. An exact match is most efficient, but if you care caching searches
 * of a term like fish->trout and fish->salmon you can delete all cached searches
 * relating to "fish" or "salmon" if data is updated.
 * 
 * 2. Caches are compressed using gzip, as the assumption is that CPU operations
 * are faster than disk operations
 * 
 * To make "secure" you can append the PHPSESSID to the hash - 
 * but this is not good on large sites, which is why Drupal has not caching for
 * logged in users!
 * 
 * 
 * @example MyClass extends SO_CA and you can then call cache functions
 * with
 * 
 * @author  Jonathan Hendler 
 * @package SONIA
 * @license GPLv2 
 * @version .3 beta
 * @date   2/20/2007
 * 
 * 
 * */

class SO_CA {
  /**
   * signal for checking cache table existance
   * */
  private $table_check = false;

  public function getCacheSize() {
    return db_result(db_query("SELECT COUNT(hash) as c FROM {cache_semantic_search}"));
  }
  
  /**
   * predict the cache size and store the search array that would be used
   * to generate searches
   * */
  public function getPredictedCacheSize(& $search) {
    $stats = theme("semantic_search_search", $search, true);
    $search_array = array ();
    $this->cacheSizeHelper($stats, $search_array);
    return $this->parseSearchArrayForSize($search_array);
  }
  
  /**
   * this is a "combination" not a permuation
   * which is n!/k!(n-k)!
   * http://en.wikipedia.org/wiki/Combination
   * */
  private static function parseSearchArrayForSize(&$search_array){
    $count = 1;
    $keys = array();
    $dkeys = array();
    $simplecount = 0;
    $k_factorial = 6; //(3*2*1)
    $k2_factorial = 2; //(2*1)
    $k_factorials = array();
    $choose_two = array();
    
    foreach($search_array as $set){
      $tkey = $set[0]['predicate'];
      if(in_array($tkey,$keys)){
        $offset = (in_array($tkey,$dkeys)) ? 2: 1; 
        $dkeys[] = $tkey;
        $setcount = count($set)-$offset;
        if (!isset($k_factorials[$tkey])){
          $k_factorials[$tkey] = $k_factorial;
          $setcount = $setcount/$k_factorial;
          $choose_two[$tkey] = (count($set) * (count($set) -1))/$k2_factorial;
          $simplecount += $setcount;
        }
      } else {
        $setcount = count($set);
        $simplecount += $setcount;
      }
      $keys[] = $tkey;
      $count = ($count * $setcount);
    }
    
    //do choose twos
    //this will need some work
    foreach($search_array as $set){
      $tkey = $set[0]['predicate'];
      if(!isset($choose_two[$tkey])){
        foreach ($choose_two as $choosekey => $amount){
          $count += $amount * count($set);
        }
      }
    }
    
    $count = $count + $simplecount;
    $count = ceil($count);
    return $count;
  }
  
   /**
   * get the search array and start looping through it
   * 
   * threads would be nice
   * */
  public function startBuildingCache(& $search) {
    $stats = theme("semantic_search_search", $search, true);
    $search_array = array ();
    $this->cacheSizeHelper($stats, $search_array);
    $completed = $this->getCache('SEMANTIC_SEARCH_CACHE_COMPLETED');
    $psize = $this->getCache('SEMANTIC_SEARCH_CACHE_P_SIZE');
    
    $keys = array();
    $dkeys = array();
    $count = $completed -1;
    $set_count = count($search_array);
    $this->search = $search;
    foreach($search_array as $k1=> $set){
      $donekeys[] = $k1;
      //constantly check if the stop signal has been sent
      $started = $this->getCache('SEMANTIC_SEARCH_CACHE_STARTED');
      if (!$started) return;
      //the array has duplicate predicates and for the top level
      // these can be avoided 
      //NTS index 0 may only be true for p_value queries
      $tkey = $set[0]['predicate'];
      
      if (!in_array($tkey,$keys)){
        foreach ($set as $subquery){
          $search->sonia->sonia_search_array  = array($subquery);
          $search->sonia->setSearchResults();
          $count++;
          //debugging
          //print ' <br /> depth '.count($subquery).' ' .count($this->search->sonia->search_results);
          $this->putCache('SEMANTIC_SEARCH_CACHE_COMPLETED',$count);
          //now iterate through all the OTHER permutations
          $this->recurseSearchArray($search_array,$donekeys, $search->sonia->sonia_search_array , $count);
        }
      }
//      if(in_array($tkey,$keys)){
//        $offset = (in_array($tkey,$dkeys)) ? 2: 1;
//        $dkeys[] = $tkey;
//        $setcount = count($set)-$offset;
//      } else {
//        $setcount = count($set);
//      }
      $keys[] = $tkey;
    }
    
    //signal that it's over
    
    $this->clearCache('SEMANTIC_SEARCH_CACHE_STARTED');  
    return $search_array;
  }
  
  /**
   * recursive function to fetch out the rest of the search array
   * */
  private function recurseSearchArray(
    &$search_array, 
    &$donekeys, 
    $subquery, 
    &$count,
    $donekeys2 = null){
    //dbg($subquery,'$subquery');
    if ($donekeys2 == null) $donekeys2 = array();
    
    foreach($search_array as $k2=>$set2){
      if (!in_array($k2, $donekeys) && !in_array($k2, $donekeys2)) {
        foreach ($set2 as $subquery2){
          $donekeys2[] = $k2;
          //check if the stop signal has been sent
          $started = $this->getCache('SEMANTIC_SEARCH_CACHE_STARTED');
          if (!$started) return;
          
          if ($this->checkNoSearchDupes($subquery, $subquery2 )){
            $tsubquery = $subquery;
            $tsubquery[] = $subquery2;
            $this->search->sonia->sonia_search_array  = $tsubquery;
            $this->search->sonia->setSearchResults();
            $count++;
            
            //print ' <br /> depth '.count($tsubquery).' ' .count($this->search->sonia->search_results);
            $this->putCache('SEMANTIC_SEARCH_CACHE_COMPLETED',$count);
            $this->recurseSearchArray($search_array,$donekeys, $tsubquery, $count,$donekeys2);
          }
        }
      }
      
    }
  }
  
  /**
   * prevents duplicate facets in a single search  
   * (basically creates the n-1 needed for correct permutations and combinations)
   * */
  private function checkNoSearchDupes(&$subquery, &$subquery2 ){
     $o = $subquery2['object'];
     $p = $subquery2['predicate'];
     foreach($subquery as $sq){
       $o2   = $sq['object'];
       $p2   = $sq['predicate'];
       if ($o == $o2 && $p == $p2) return false;
     }
     return true;
  }
  
  private function cacheSizeHelper(& $array, & $search_array) {
    /**
     * [search_num][fnum][subject] => 
            [predicate] => http://n4/namespaces/semantic_search/Property#field_types_of_spaces_included
            [object] => Gallery
            [literal_object] => 1
            [literal_type] => exact
     * */
     //dbg($array,'whats this?');
    if (!is_array($array)){
      
      $array = array();
    }
     
    foreach ($array as $key => $value) {
      if ($this->isKey($key)) {
        $temp_array = $temp_array_a = $temp_array_b = array();
        $temp_count = 0;
        $sub_array = array();
        $sub_array['literal_object'] = true;
        $sub_array['literal_type'] = 'exact';
        $sub_array['subject'] = '';
        
        
        $options = $value['#options'];
        $uri = $this->getURI($key);
        if (!empty ($uri)) {
          $sub_array['predicate'] = $uri;
          
          if ($this->isMultiple($key)) {
            foreach ($options as $k => $v) {
              if ($k != '0'){
                $sub_array['object'] = SONIA_H :: decodeURI($k);
                $temp_array[] = $sub_array;
                $temp_array_a[] = $sub_array;
                $temp_array_b[] = $sub_array;
              }
            }
          } else {
            foreach ($options as $k => $v) {
              if($k != '0'){
                 $sub_array['object'] = SONIA_H :: decodeURI($k);
                 $temp_array[] = $sub_array;
                 
              }
            }
          }
        }
        
         $search_array[] = $temp_array;
         if (!empty($temp_array_a))  $search_array[] = $temp_array_a;
         if (!empty($temp_array_b))  $search_array[] = $temp_array_b;
        
      }
     
      
      //if this isn't the right level, recurse further down
      if (is_array($value)) {
        $this->cacheSizeHelper($value, $search_array);
      }

    }
  }

  private function isKey($key) {
    return strstr($key, 'p_value'); // || strstr($key,'p_value') ;
  }

  private function isMultiple($key) {
    return strstr($key, '][][');
  }

  private function getURI($key) {
    $match = array ();
    if (preg_match('/\[(.*)/i', $key, $match))
    $rval = str_replace('[','',$match[0]);
    $rval = str_replace(']','',$rval);
    return SONIA_H :: decodeURI($rval);
  }

  ////////////////////////////////////////////////////////////////////////////
  /**
   * place a variable in cache
   * 
   * @param $varname
   * @param $var
   * @param string $varname the name of the variable
   * @param boolean $persist should the cache exist outside of the session and the user
   * 
   * @todo add switch to determine if the object needs to be serialized
   * @todo this is global cache only
   * @todo are there problems with gzip data in DB?
   * */
  ////////////////////////////////////////////////////////////////////////////
  public function putCache($varname, $var, $persist = true, $array = false) {
    if ($varname == false)
      return false;
    $this->_preCache();

    $this->encode($var, $array);

    if ($persist) {
      db_lock_table('cache_semantic_search');
      @ db_query("UPDATE {cache_semantic_search} SET cash= %b WHERE hash= '%s'", $var, $varname);
      if (!db_affected_rows())
        @ db_query("INSERT INTO {cache_semantic_search} (hash,cash) VALUES ('%s', %b )", $varname, $var);
      if (!db_affected_rows()) {
        //try again, without error suppression
        db_query("INSERT INTO {cache_semantic_search} (hash,cash) VALUES ('%s', %b )", $varname, $var);
      }
      db_unlock_tables();
    } else { //use standard user specific session 
      $_SESSION[$varname] = $var;
    }
    $this->_postCache();
  }

  ////////////////////////////////////////////////////////////////////////////
  /**
  * get a variable from cache
  * @param string $varname the name of the variable
  * @param boolean $persist should the cache exist outside of the session_hash
  * */
  ////////////////////////////////////////////////////////////////////////////
  public function getCache($varname, $persist = true, $array = false) {
    if ($varname == false)
      return false;
    $this->_preCache();

    if ($persist) {
      $cache = db_fetch_object(db_query("SELECT cash FROM {cache_semantic_search} WHERE hash= '%s' ", $varname));
      if (db_error()) {
      }
      $this->_postCache();
      if (!empty ($cache->cash)) {
        $this->decode($cache->cash, $array);
        return $cache->cash;
      } else {
        return null;
      }

    } else
      if (isset ($_SESSION[$varname])) {
        $this->_postCache();
        return $this->decode($_SESSION[$varname], $array);
      } else {
        $this->_postCache();
        return null;
      }

  }

  /**
   * hash generating function
   * 
   * the array is not simply serialized. The hashing function
   * searches creates 
   * 
   *  The array must be less than 7 if use keys is true (default), and 15 
   *  if false because the key field in the DB is max 512
   * 
   * @param mixed, either an array of key values, or a value
   * @return mixed (string|boolean) 
   * */
  static function gh($key, $prefix = null, $use_keys = false) {
    if (is_array($key)) {
      $size = count($key);
      $returnstring = '';
      $tarray = array ();
      $count = 0;

      foreach ($key as $k => $v) {
        if (is_array($v)) {
          error_handler(1, "SO_CA::gh param \$key must not be a multi dimensional array", "semantic_search/contrib/SONIA/ext/cache/ffcache.sonia.inc", 140);
          return false;
        }
        if ($use_keys)
          $tarray[$count] = md5($k) . '-';
        if ($use_keys)
          $tarray[$count] .= md5($v) . '-';
        ++ $count;
      }
      arsort($tarray);

      foreach ($tarray as $x) {
        $returnstring .= $x;
        if (strlen($returnstring) > SONIA_CACHE_KEY_LENGTH) {

          $returnstring = 'rehashed' . md5($returnstring);
        }
      }

    } else {
      $returnstring = md5($key);
    }

    return $prefix . $returnstring;
  }

  /**
  * called priort to cache function exectution
  * */
  private function _preCache() {

  }

  /**
   * called after cache function exectution
   * */
  private function _postCache() {
  }

  /**
  * erase a cache
  * */
  public function clearCache($key = null, $persist = true) {

    $this->_preCache();
    if ($persist && $key !== null) {
      //cache_clear_all($key);
      db_query(" DELETE FROM {cache_semantic_search} WHERE hash= '%s' ", $key);
    } else
      if (!$persist && $key !== null) {
        unset ($_SESSION[$key]);
      } else
        if ($persist) {
          $sql = " TRUNCATE TABLE {cache_semantic_search} ";
          $res = db_query($sql);
        }

    $this->_postCache();
  }

  /**
   * encodes string for DB
   * detects arrays
   * 
   * */
  private function encode(& $var, $array) {
    if ($array)
      $var = serialize($var);
    if (function_exists('gzcompress')) {
      //$complevel = (strlen($var) > 2) ? 1 : 9;
      //assumes that it is a string under 2MB
      $var = gzcompress($var, 9);
      $var = addslashes($var);

    } else {
      $var = base64_encode($var);
    }

  }

  /**
   * Decodes value from db
   * 
   * @return mixed
   * */
  private function decode(& $var, $array) {

    if (function_exists('gzuncompress')) {
      $var = db_decode_blob($var);
      $var = stripslashes($var);
      $var = gzuncompress($var);

    } else {
      $var = base64_decode($var);
    }

    if ($array) {
      $var = unserialize($var);
    }
  }

  /**
   * function to determine if a string is a serialized array
   * DEPRECATED - if string is large, this is inneffecient
   *  */
  private function isa(& $val) {
    $str = 's';
    $array = 'a';
    $integer = 'i';
    $any = '[^}]*?';
    $count = '\d+';
    $content = '"(?:\\\";|.)*?";';
    $open_tag = '\{';
    $close_tag = '\}';
    $parameter = "($str|$array|$integer|$any):($count)" . "(?:[:]($open_tag|$content)|[;])";
    $preg = "/$parameter|($close_tag)/";
    return preg_match($preg, $val, $match);
  }
} //end class
