<?php

/**
 * SONIA
 * Simple Ontology aNd Inference Algorithms
 * 
 * The SONIA Store API provides an internal API for SONIA API to interact with a 
 * an RDF store, Triple Store or other store
 * 
 * 
 * @todo full CRUD operations
 * 
 *
 * 
 * @author  Jonathan Hendler 
 * @package SONIA
 * @license GPLv2
 * @license copyright 2006 KBCMS LLC all other uses
 * @version .03 alpha
 * @date    4/19/2006
 * 
 * 
 * */
interface SONIAStoreAPI {

    //SYSTEM
    //--------------------------------------------------------------------------
    /**
     * Parse and check params set to create/ login to the store.
     * 
     * @param array $args an array of arguments needed for the store. 
     * eg. for mysql, db name, username, password etc 
     *  
     * */
    //--------------------------------------------------------------------------
    public function init($args);
    
    //--------------------------------------------------------------------------
    /**
     * Via a Cron type functionality, perform maintainence tasks
     * 
     * @param array $args an array of arguments needed for the store. 
     * eg. for sync, export, cache, save, etc
     *  
     * */
    //--------------------------------------------------------------------------
    public function maintain($args);

    //RETRIEVAL
    //--------------------------------------------------------------------------
    /**
    * basic function to confirm if a statement exists,
    * used to determine if a class, or paramter has already been created
    * 
    * @param string uri subject
    * @param string uri predicate
    * @param string uri object
    * @param boolean - see if the subject (one URI) exists anywhere in a statement
    * 
    * @return boolean does the statement exist
    * */
    //--------------------------------------------------------------------------
    public function statementExists($s, $p = null, $o = null, $any = true);

    //--------------------------------------------------------------------------
    /**
     * Convert array to SPARQL
     * 
     * @param array $query, the statement sent via SONIA
     * 
     * @return string $convertedQuery 
     * @see SONIA :: UI_query()
     * @see query();
     * */
    //--------------------------------------------------------------------------
    public function convertQuery($query);

    //--------------------------------------------------------------------------
    /**
     * Parse the query and send a generalized query to the store
     * 
     * @see convertQuery();
     * @see @see SONIA :: UI_query()
     * 
     * @return array results and result codes to be used by SONIA
     * 
     * */
    //--------------------------------------------------------------------------
    public function query($convertedQuery, $return_type = 'array');
    //DATA CREATION

    //--------------------------------------------------------------------------
    /**
     * takes a stores search results and parses it to sonia search results.
     * @param array results from some format
     * @return array a standard SONIA results array
     * */
    //--------------------------------------------------------------------------
    public function parseStoreToSonia(&$results);

    //--------------------------------------------------------------------------
    /**
    @param
    @return int sid the sid just created
    @todo add user and group permissions management
    @todo add model concept
    */
    //--------------------------------------------------------------------------
    function createStatement($subject, $predicate, $object, $object_literal = false, $dataType = 'string', $owner = null, $group = null, $default_ns = null);

    function updateStatement($subject, $predicate, $object, $new_subject, $new_predicate, $new_object, $object_literal = false, $dataType = 'string', $owner = null, $group = null, $default_ns = null);

    function deleteStatement($subject, $predicate, $object, $object_literal = false, $dataType = 'string', $owner = null, $group = null, $default_ns = null);

    //--------------------------------------------------------------------------
    /**
     * 
     * Creates an RDFS:Class
     * 
     * USAGE:
     * 
     * @param string class name is the name of the class
     * 
     * @return void
     * 
     * */
    //--------------------------------------------------------------------------
    public function createClass($label, $super_categories, $id = null);

    //--------------------------------------------------------------------------
    /**
     * 
     * sets one RDFS:Class to be a subclass of another
     * 
     * USAGE:
     * 
     * @param string URI of the class
     * @param mixed URI of the superclasses
     * 
     * @return void
     * 
     * */
    //--------------------------------------------------------------------------
    public function setSubclass($class, $superclasses);

    //--------------------------------------------------------------------------
    /**
     * 
     * creates an RDF:Property for an RDFS:Class 
     * 
     * USAGE:
     * 
     * @param string start name or id of the start category 
     * 
     * @return void
     * 
     * */
    //--------------------------------------------------------------------------
    public function addPropertyToClass($class, $property);

    //--------------------------------------------------------------------------
    /**
     * 
     * places facet and its value in an existing class. Same as an RDF Literal
     * 
     * USAGE:
     * 
     * @param string $object_name_or_id the reference to the object
     * @param string $facetType the identifier for the property
     * @param string $data the data
     * @param string $dataType int, string, etc , not enforced
     * 
     * 
     * @return array a multi-deminsional array expressing the hierarchy of 
     * categories
     * 
     * @see createObject()
     * */
    //--------------------------------------------------------------------------
    public function setObjectProperty($object_name_or_id, $property, $data, $dataType);

    //--------------------------------------------------------------------------
    /**
    * gets all the categories/classes and counts available for a search
    * 
    * @param string in the list of subjects
    * */
    //--------------------------------------------------------------------------
    public function getAllCategories(& $count, $startclass = 'Space', $in = null);

    //-------------------------------------------------------------------------- 
    /**
    * get a namespace id for a namespace
    * 
    * 
    * @param string ns the namespace
    * @return int nid the primary key of the name space
    * 
    * @todo owner, group
    * 
    * 
    * */
    //--------------------------------------------------------------------------
    public function getNSID($ns);

    //--------------------------------------------------------------------------
    /**
     * 
     * Retrieves the number of objects in the category
     * 
     * USAGE:
     * 
     * @param string $category_name_or_id name or id of the start category 
     * 
     * @return int count
     * 
     * @see query()
     * */
    //--------------------------------------------------------------------------
    public function getCategoryCount($category_name_or_id);

    //--------------------------------------------------------------------------
    /**
    * gets all properties per class
    * 
    * @todo does not handle inference!
    * @param string in the list of subjects
    * */
    //--------------------------------------------------------------------------
    public function getAllClassProperties(& $count, $startclass);

    //--------------------------------------------------------------------------
    /**
    * gets all the categories/classes and counts available for a search
    * 
    * @param string count, the number of results
    * @see getAllClassProperties
    * */
    //--------------------------------------------------------------------------
    public function getAllProperties(& $count);

    //--------------------------------------------------------------------------
    /**
     * 
     * Retrieves the number of objects having a facet 
     * 
     * USAGE:
     * 
     * @param string $property_name_or_id name or idof the property 
     * 
     * @return int count
     * 
     * @see query()
     * */
    //--------------------------------------------------------------------------
    public function getPropertyCount($property_name_or_id);

    //--------------------------------------------------------------------------
    /**
     * import RDF into the rdf store 
     * 
     * @todo validate the RDF
     * @param string $iri_model the name of the model
     * @param string $file the location of the file, can be any uri PHP allows http:// file:/// or just a path
     * */
    //--------------------------------------------------------------------------
    public function importRDF($iri_model, $file);

    //--------------------------------------------------------------------------
    /**
     * eport RDF to a string
     * 
     * @todo use iri model
     * @param string $iri_model the name of the model
     * @return string
     * */
    //--------------------------------------------------------------------------
    public function exportRDF($iri_model);

} //end interface
