<?php

/**
 * Provides base functions that mostly to do with data editing and creation,
 * and those not directly implementing the API such as
 * helper functions, array handling functions, and other algorithms,
 * or functions that are part of the API but only depend on other elements  
 * 
 * @package SONIA
 * */
abstract class SONIABase {

    /**
     * wraps the store function for updating a property
     * if $old_value is null it will erase all those statements 
     * 
     * @todo explore issues with multiple cardinality and filters and access rights
     * */
    public function updatePropertyOfObject($subject, $property, $old_value, $new_value) {
        //quick safety checks
        if (empty ($subject) || empty ($property))
            return;

        return $this->store->updateStatement($subject, $property, $old_value, $subject, $property, $new_value, true);
    }

//    function _getCategoryInternalCount($category_id) {
//        return $this->getSearchPropertyCount($category_id, SONIA_INSTANCE);
//    }

    public function getAllClassProperties(& $count, $startclass) {
        return $this->store->getAllClassProperties($count, $startclass);
    }

    public function addPropertyToClass($class, $property) {
        return $this->store->addPropertyToClass($class, $property);
    }

    public function createClass($classname, $superclasses, $id = null) {
        return $this->store->createClass($classname, $superclasses, $id);
    }

    public function createProperty($property_name, $classes = null, $id = null) {
        return $this->store->createProperty($property_name, $classes, $id);
    }

    public function updateProperty($property, $property_name, $classes) {
        return $this->store->updateProperty($property, $property_name, $classes);
    }

    public function createComment($class, $comment) {
        return $this->store->createStatement($class, SONIA_RDFS_CREATE_COMMENT, $comment, true);
    }

    //--------------------------------------------------------------------------
    /**
     * remove a property from the definistion of a class.
     * Data on specific objects are preserved
     * 
     * @param string class - the class being defined
     * @param string property - the property which belongs to the class
     * */
    //--------------------------------------------------------------------------    
    public function deletePropertyFromClass($class, $property) {
        return $this->store->deleteStatement($property, SONIA_RDF_CREATE_PROPERTY_OF, $class);
    }

    //--------------------------------------------------------------------------
    /**
     * remove a property from the definistion of a class.
     * Data on specific objects are preserved
     * 
     * @param string property - the property which belongs to the class
     * */
    //--------------------------------------------------------------------------    
    public function deleteProperty($property, $class = null) {
        if ($property == '')
            return;
        if ($class == '') {
            $this->deleteAllTracesOfStatement($property);
        } else {
            //remove values?
            $this->store->deleteStatement($class, $property, null);
            //remove domain
            $this->store->deleteStatement($property, null, $class);
        }
    }

    //--------------------------------------------------------------------------
    /**
     * remove a property from the definistion of a class.
     * Data on specific objects are preserved
     * 
     * @param string uri
     * */
    //--------------------------------------------------------------------------    
    public function deleteAllTracesOfStatement($uri) {
        $this->store->deleteStatement($uri, null, null);
        $this->store->deleteStatement(null, $uri, null);
        $this->store->deleteStatement(null, null, $uri);
    }

    //--------------------------------------------------------------------------
    /**
     * 
     * creates an instance of an RDFS:Class 
     * 
     * USAGE:
     * 
     * @param string $object_name_or_id the reference to the object
     * @param array $object_category_list list of categories to which the object belongs
     * @param array $object_param_list list of parameters and there values,datatypes, literal or not
     * 
     * @todo throw error if no label
     * @todo add datatype enforcement and entry
     * @todo check to see that the class exists
     * 
     * @return void
     * 
     * @see setObjectFacet()
     * */
    //--------------------------------------------------------------------------
    public function updateObject($id, $label, $classes, $parameters) {
        $id_iri = $id;
        if ($label == '' || $label == null || empty ($id))
            return;
        $this->updateLabel($id_iri, $label);

        if (isset ($parameters['images']) && is_array($parameters['images'])) {
            foreach ($parameters['images'] as $image) {
                //create image and add to object as general file type

                $image_uri = $this->createImage($id_iri, $label, ' ', $image['filename'], $image['filesize'], $image['filepath']);

                drupal_set_message('SEMANTIC_SEARCH: added image');
                //adds direct reference here
                $this->store->updateStatement($id_iri, $image['uri'], null, $id_iri, $image['uri'], $image_uri);
            }
        }
        
        //TODO add class based datatype check
        foreach ($parameters as $key => $parameter) {
            $old_value = (isset ($parameter['old_value'])) ? $parameter['old_value'] : null;
            if ($key != 'images' && ($parameter['uri'] != SONIA_RDF_CREATE_TYPE)) {
                $this->updatePropertyOfObject($id_iri, $parameter['uri'], $old_value, $parameter['value']);
                
                if (isset($tester[$id_iri][$parameter['uri']])){
                  dbg("duplcate of $id_iri -> {$parameter['uri']} ");
                }
                else
                {
                  $tester[$id_iri][$parameter['uri']] =1;
                }
                
            }
        }

        if (!empty ($classes)) {
            //this needs to add all new classes, remove all old ones in two steps
            $this->store->deleteStatement($id_iri, SONIA_RDF_CREATE_TYPE, null);

            foreach ($classes as $class) {
                $this->store->createStatement($id_iri, SONIA_RDF_CREATE_TYPE, $class);
            }
        } else {
            dbg($classes, 'classes were empty!!!!!!!!!!');
        }
        //create property
        //TODO not needed if duplicate
        return $id_iri;
    }

    /**
     * set a property as an image type
     * */
    public function setAsImageType($property_uri) {
        return $this->store->createStatement($property_uri, SONIA_RDF_CREATE_TYPE, SONIA_FILE_TYPE_URI);
    }
    /*
     * 
      define('SONIA_HAS_IMAGE','sonia:hasImage');
      define('SONIA_CREATE_HAS_IMAGE',SONIA_BASE_IRI.'#hasImage');
      
      define('SONIA_HAS_FILE','sonia:hasFile');
      define('SONIA_CREATE_HAS_FILE',SONIA_BASE_IRI.'#hasFile');
     
     
     
     */

    //--------------------------------------------------------------------------    
    /**
     * 
     * 
     * @param string $object_uri object the image is attached to
     * @param string $label
     * @param string $description, description of the image
     * @param string $filename
     * @param string $filesize
     * @param string $path, absolute URI to retrieve the image itself
     * @param string $id the ID identifying the file
     * 
     * @return string $uri the new object URI
     * */
    //--------------------------------------------------------------------------
    public function createImage($object_uri, $label, $description, $filename, $filesize, $path, $id = null) {
        return $this->createFile($object_uri, $label, $description, $filename, $filesize, $path, SONIA_CREATE_HAS_IMAGE, $id);
    }

    //--------------------------------------------------------------------------
    /**
     * @param string $object_uri object the image is attached to
     * @param string $label
     * @param string $description, description of the image
     * @param string $filename
     * @param string $filesize
     * @param string $path, absolute URI to retrieve the image itself
     * @param string URI type of file
     * @param string $id an ID identifying the file
     * */
    //--------------------------------------------------------------------------
    public function createFile($object_uri, $label, $description, $filename, $filesize, $path, $rdf_type = null, $id = null) {

        if (empty ($file_uri)) {
            $id = $this->store->generateFileAutoIncrementField();
            $file_uri = SONIA_FILE_ID_URI . $id;
        } else {
            $file_uri = SONIA_FILE_ID_URI . $id;
        }

        if (empty ($rdf_type))
            $rdf_type = SONIA_CREATE_HAS_FILE;
        //add file uri to to object URI with $rdf_type as predicate  

        //make this a type of file

        if (!empty ($object_uri)) {
            $this->store->createStatement($object_uri, $rdf_type, $file_uri);
        }

        //create label
        $this->store->createStatement($file_uri, SONIA_RDFS_CREATE_LABEL, $label, true);
        //create description
        $this->createComment($file_uri, $description);

        //define('SONIA_CREATE_FILENAME',SONIA_BASE_IRI.'#soniaFileName');
        //define('SONIA_CREATE_FILESIZE',SONIA_BASE_IRI.'#soniaFileSize');
        //define('SONIA_CREATE_FILEPATH',SONIA_BASE_IRI.'#soniaFilePath');

        //create filename
        $this->store->createStatement($file_uri, SONIA_CREATE_FILENAME, $filename, true);

        $this->store->createStatement($file_uri, SONIA_CREATE_FILESIZE, $filesize, true);

        $this->store->createStatement($file_uri, SONIA_CREATE_FILEPATH, $path, true);

        $this->store->createStatement($file_uri, SONIA_RDF_CREATE_TYPE, SONIA_FILE_CLASS_URI);

        return $file_uri;
    }

    //--------------------------------------------------------------------------    
    /**
     * @todo
     * */
    //--------------------------------------------------------------------------
    public function updateImage() {

    }

    //--------------------------------------------------------------------------
    /**
     * @todo
     * */
    //--------------------------------------------------------------------------
    public function updateFile() {

    }

    //--------------------------------------------------------------------------    
    /**
     * 
     * */
    //--------------------------------------------------------------------------
    public function deleteImage($object_uri, $image_uri) {

    }

    //--------------------------------------------------------------------------
    /**
     * 
     * */
    //--------------------------------------------------------------------------
    public function deleteFile($object_uri, $image_uri, $type) {

    }

    //--------------------------------------------------------------------------
    /**
     * 
     * creates an instance of an RDFS:Class 
     * 
     * USAGE:
     * 
     * @param string $object_name_or_id the reference to the object
     * @param array $object_category_list list of categories to which the object belongs
     * @param array $object_param_list list of parameters and there values,datatypes, literal or not
     * 
     * @todo throw error if no label
     * @todo add datatype enforcement and entry
     * @todo check to see that the class exists
     * 
     * @return void
     * 
     * @see setObjectFacet()
     * */
    //--------------------------------------------------------------------------
    public function createObject($label, //TODO not needed?
    $classes, $parameters, $id = null) {
        
        if ($label == '' || $label == null)
            return;

        if ($id == null) {
            //create ID 
            //TODO check that it doesn' exist
            $id = $this->store->generateObjectAutoIncrementField();
            $id_iri = SONIA_OBJECT_ID_URI . $id;
        } else {
            //assume the id is a full IRI
            //URIExists($s)
            $id_iri = $id;
        }
        
        //create label
        $this->store->createStatement($id_iri, SONIA_RDFS_CREATE_LABEL, $label, true);

        if (is_array($classes)) {
            foreach ($classes as $class) {
              //SONIA_RDF_CREATE_ABOUT
              //$this->store->createStatement($class, SONIA_RDF_CREATE_ABOUT, $id_iri);
                $this->store->createStatement($id_iri, SONIA_RDF_CREATE_TYPE, $class);
            }
        } else
            if (!empty ($classes)) {
              //$this->store->createStatement($classes, SONIA_RDF_CREATE_ABOUT, $id_iri);
              $this->store->createStatement($id_iri, SONIA_RDF_CREATE_TYPE, $classes);
                //$this->store->createStatement($id_iri, SONIA_RDF_CREATE_TYPE, $classes);
            }

//        if (isset ($parameters['images']) && is_array($parameters['images'])) {
//            foreach ($parameters['images'] as $image) {
//                //create image and add to object as general file type
//                $image_uri = $this->createImage($id_iri, $label, ' ', $image['filename'], $image['filesize'], $image['filepath']);
//
//                //adds direct reference here
//                $this->store->createStatement($id_iri, $image['uri'], $image_uri, true);
//            }
//        }
   //     dbg($parameters,'$parameters');
        
        //TODO add class based datatype check
        foreach ($parameters as $key => $parameter) {
            //dbg($parameter,($parameter['value'] != '') .'--'.   ('images' != $key) . " {$parameter['value']} != '' and $key != 'images'");
            if (($parameter['value'] != '') &&  ('images' != $key)) {
                $this->store->createStatement($id_iri, $parameter['uri'], $parameter['value'], true);
            }
        }

        //create property
        //TODO not needed if duplicate
        return $id_iri;
    }

    public function updateComment($class, $comment) {
        return $this->store->updateStatement($class, SONIA_RDFS_CREATE_COMMENT, null, $class, SONIA_RDFS_CREATE_COMMENT, $comment, true);
    }

    public function updateLabel($class, $label) {
        return $this->store->updateStatement($class, SONIA_RDFS_CREATE_LABEL, null, $class, SONIA_RDFS_CREATE_LABEL, $label, true);
    }

    public function deleteComment($class) {
        return $this->store->deleteStatement($class, SONIA_RDFS_CREATE_COMMENT, null);
    }

    //--------------------------------------------------------------------------
    /**
     * 
     * Creates an RDFS:Class
     * 
     * USAGE:
     * @param string $id is the ID or URI of the class
     * @param string class name is the name of the class
     * @param array $superclasses are the set of superclasses
     * @todo replace IDs
     * 
     * @return void
     * 
     * */
    //--------------------------------------------------------------------------
    public function updateClass($id, $classname, $superclasses = null) {
        //assume the id is a full IRI
        $id_iri = $id;
        //dbg($superclasses,'update class '.$classname);
        if (empty ($superclasses)) {
            //todo make this the root class
            $superclasses = array ();
        }

        //TODO if a class has multiple labels this will be be bad
        $this->store->updateStatement($id_iri, SONIA_RDFS_CREATE_LABEL, null, $id_iri, SONIA_RDFS_CREATE_LABEL, $classname, true);

        if (!empty ($superclasses)) {
            //remove ALL previous subclass associations
            $this->store->deleteStatement($id_iri, SONIA_RDFS_CREATE_SUBCLASS, null);

            foreach ($superclasses as $superclass) {
                $this->store->createStatement($id_iri, SONIA_RDFS_CREATE_SUBCLASS, $superclass);
            }
        }

        $this->store->createStatement(
            $id_iri,
            SONIA_RDF_CREATE_TYPE,
            SONIA_RDFS_CREATE_CLASS_TYPE 
        );
        return $id_iri;
    }

    public function getClasses($object) {
        return $this->store->getClasses($object);
    }

    //helper
    public function getReservedPredicates() {
        return $this->reserved_predicates;
    }

    /**
    * add an aditional label to any class, object, or other items for the purposses search
    * */
    public function addTag() {

    }

    public function removePropertyFromObject($subject, $property, $old_value, $literal = true) {
        if (empty ($subject) || empty ($property))
            return;
        return $this->store->deleteStatement($subject, $property, $old_value, $literal);
    }

    public function addPropertyToObject($subject, $property, $value, $literal = true) {
        if (empty ($subject) || empty ($property))
            return;
        return $this->store->createStatement($subject, $property, $value, $literal);
    }

    //TODO
    public function createCategory($label, $id, $parents) {
        //create id
        $this->store->createStatement($subject, $property, $value, false);

        //create label
        return $this->store->createStatement($subject, $property, $label, true);
    }

    /**
     * Delete a class
     * 
     * Do not delete objects. Does not delete all children as this would be unneccesary 
     * in RDF, however in Drupal, they choose to delete all inhereted classes
     * 
     * @param string $uri
     * @param boolean delete subclass, whether or not to delete a classess subclasses
     * @param boolean $delete_all_objects_related delete all related objects
     * 
     * @todo delete all releated objects
     * */
    public function deleteClass($uri, $delete_subclasses = true, $delete_all_objects_related = false) {
        if ($uri == '' || $uri == null)
            return false;

        if ($delete_all_objects_related) {
            //get all objects of this class - or subclasses
            $sonia_search_array[] = array (
                'subject' => null,
                'predicate' => SONIA_RDF_CREATE_TYPE,
                'object' => $uri,
                'literal_object' => false,
                'category_query' => true
            );

            //this get's all objects of class 2
            $results = $this->query(0, $sonia_search_array, null, null, false, '', '');
            //delete all the objects related
            if (is_array($results)){
              foreach ($results as $turi => $data) {
                  $this->deleteObject($turi);
              }
            }
        }

        //get the direct subclasses of this class
        $subclasses = SONIA_H :: flattenHieararchy($this->store->getAllCategoriesRecursive($uri, 0, true));
        //dbg($subclasses,'deleteiing subclasses of '. $uri);

        //del
        $this->store->deleteStatement($uri, null, null);
        $this->store->deleteStatement(null, $uri, null);
        $this->store->deleteStatement(null, null, $uri);

        if ($delete_subclasses) {

            foreach ($subclasses as $subclass) {
                $this->store->deleteStatement($subclass, null, null);
                $this->store->deleteStatement(null, $subclass, null);
                $this->store->deleteStatement(null, null, $subclass);
            }

        } //get the direct superclasses of this class
        else
            if (!$delete_subclasses) {
                //get the direct subclasses of this class

                //delete all references to the class

                //add the superclasses to the subclasses
            }

    }

    /**
     * Removes all references to a particular object
     * 
     * @todo Needs security.
     * @todo could be made more efficient with a single SPARQL/SQL statement, if SPARQL changes
     * 
     * */
    public function deleteObject($subject) {
        if ($subject == '')
            return;
        return $this->deleteAll($subject);
    }

    public function deleteAll($subject) {
        //TODO SECURITY make sure this is really just deleting 
        //insance of an object!!!
        if (empty ($subject))
            return false;
        $this->store->deleteStatement(null, null, $subject);

        $this->store->deleteStatement(null, $subject, null);

        return $this->store->deleteStatement($subject, null, null);
    }

    public function _GetSONIADebugText() {
        return $this->sonia->debug_text;
    }

    //-------------------------------------------------------------------------- 
    /**
    * Will search the request and parse to provide updates to the data
    * 
    * @see doCRUD()
    * @todo add user permissions and managment
    * */
    //--------------------------------------------------------------------------
    public function parseCRUD() {

        if ($this->io->doCRUD()) {
            $values = $this->io->getRequestValues();
            if (isset ($values[SONIA_FORM_REQUEST_PREFIX]))
                $values = $values[SONIA_FORM_REQUEST_PREFIX];

            switch ($values[SONIA_CRUD_ACTION_SIGNAL]) {
                case SONIA_CRUD_UPDATE_SIGNAL :

                    if (is_array($values) && !empty ($values)) {
                        foreach ($values as $k => $v) {
                            //TODO why are some of these sent as bad Predicates
                            if ($this->notCrudKey($k) && strlen($v[SONIA_CRUD_PREDICATE]) > 1 //HACK extra protection
                            && $this->io->notReservedURLKey($k)) {
                                $this->updatePropertyOfObject(SONIA_H :: decodeURI($values[SONIA_CRUD_SUBJECT]), SONIA_H :: decodeURI($v[SONIA_CRUD_PREDICATE]),
                                //TODO should not be decoded... but preped for DB //security etc
                                SONIA_H :: decodeURI($v[SONIA_CRUD_OLD_OBJECT]), SONIA_H :: decodeURI($v[SONIA_CRUD_OBJECT]));
                            }
                        }
                    } else {
                        $this->updatePropertyOfObject(SONIA_H :: decodeURI($values[SONIA_CRUD_SUBJECT]), SONIA_H :: decodeURI($values[SONIA_CRUD_PREDICATE]), SONIA_H :: decodeURI($values[SONIA_CRUD_OLD_OBJECT]), SONIA_H :: decodeURI($values[SONIA_CRUD_OBJECT]));
                    }
                    break;
                case SONIA_CRUD_CREATE_SIGNAL :
                    $this->addPropertyToObject(SONIA_H :: decodeURI($values[SONIA_CRUD_SUBJECT]), SONIA_H :: decodeURI($values[SONIA_CRUD_PREDICATE]), SONIA_H :: decodeURI($values[SONIA_CRUD_OBJECT]));
                    break;

                case SONIA_CRUD_DELETE_SIGNAL :
                    $this->removePropertyFromObject(SONIA_H :: decodeURI($values[SONIA_CRUD_SUBJECT]), SONIA_H :: decodeURI($values[SONIA_CRUD_PREDICATE]), SONIA_H :: decodeURI($values[SONIA_CRUD_OLD_OBJECT]));
                    break;

                case SONIA_CRUD_DELETE_OBJECT_SIGNAL :
                    $this->removeObject(SONIA_H :: decodeURI($values[SONIA_CRUD_SUBJECT]));
                    break;
            }
        }
        //else do nothing
    }

    //--------------------------------------------------------------------------
    /**
     * provides conversion from a SONIA native query return type
     * to a string 
     * 5 base64 encoded fields, S, P, O, and OL (indicates if the O is literal), and T for type 
     * 
     * @todo use type for more than just string
     * 
     * @return string
     * */
    //--------------------------------------------------------------------------
    function parseNativeTripleToString($results) {
        $output_string = '';
        // define('SONIA_EXPORT_DELIMITER',' ||| '); 
        // define('SONIA_EXPORT_NEWLINE',' ---'."\n");
        foreach ($results as $result) {

            $output_string .= SONIA_EXPORT_NEWLINE;
        }
        return $output_string;
    }

}
