<?php

/**
 * SONIA
 * Simple Ontology and Inference Algorithms
 * 
 * The SONIA API provides a standardized API for the use of RDF stores and other RDBMS
 * solutions to provided faceted search. 
 * 
 * This is the core file of the SONIA implementation. This file requires its API 
 * and it also requires an IO object and a Store object
 *  
 * @see SONIA_IO
 * @see SONIAStoreAPI
 * 
 * @todo permissions management
 * @todo model management
 * @todo inference
 * @todo import from RDF
 * @todo export from RDF
 * @todo set distinct
 * @todo add weights to statements to allow for toplogical sorts
 * 
 * @license GPL-V2 FSF
 * @license (Commercial) 2006 KBCMS LLC
 * @author  Jonathan Hendler <[hendlerman (at) yahoo DOT com]>
 * @package SONIA
 * @version 0.3
 * 
 * 
 * */
class SONIA extends SONIASearch implements SONIAAPI {

  /**
   * array of known abbreviations for namespaces
   * */
  private $ns_table;

  /**
   * string , the variable for the default long namespace
   * */
  private $default_ns_long;

  /**
  * string , the variable for the default abv namespace
  * */
  public $default_ns_abrev;

  /**
   * total categories available for search
   * */
  public $total_category_count;

  /**
   * all categories available in search
   * */
  public $all_categories;

  /**
   * the backend RDF/Triple store
   * */
  public $store;
  /**
   * the search params stored in an associative array
   * */
  public $search_params;

  /**
   * 
   * */
  public $sonia_search_array;
  /**
   * an array of defines that may be checked in the constructor
   * */
  public $reserved_defines;

  /**
   *  an array of defines set in the namespace that should be reserved
   * and not overwritten
   * */
  public $reserved_predicates;

  /**
   * session hash is the key used to retrieve saved searches
   * */
  public $session_hash;

  /**
   * lots of debugging output
   * constant SONIA_DEBUG must be on
   * */
  public $debug_text;

  /**
   * IO handles caching, sessions, file IO 
   * */
  public $io;

  //INITIALIZATION
  //--------------------------------------------------------------------------
  /**
  * Constructor
  * 
  * */
  //--------------------------------------------------------------------------
  function __construct($store = 'sonia_native_local', $request_param = 'semantic_search', $io = 'drupal', $request = null, $session = null, $install = false) {
    //set the member var for request_param
    $this->request_param = $request_param;

    //init the store and IO
    $this->_init($store, $io);

    if (!defined('SONIA_PRINT_MESSAGE_METHOD'))
      define('SONIA_PRINT_MESSAGE_METHOD', 'drupal_set_message');
    if ($install)
      $this->store->installBackend(SONIA_PRINT_MESSAGE_METHOD);

    //init vars via IO
    $this->_setIORoutine();
    if ($this->searchSubmitted()) {
      $this->_doSearch();
    }
  }

  /**
   * A simplification of getCustomSearchParamsData().
   * Given a category/class, parameter, and an optional FILTER (see SPARQL)
   * return an array of IDs(IRIs), labels, and counts (total in search, total) which can be used
   * to construct dynamic search forms. For ranges on return values or labels use
   * getTelescopingSearchParamsData().
   * 
   * SONIA depends on a PHP memory hashtable (array) to store _all_ classes and
   * params and their labels, so this _may_ be restrictive for very large 
   * data sets.  
   * 
   * The function is basically a search operation. Return values are cached 
   * by the function unless otherwise expressed.
   * 
   * @param string startclass the startclass for the category
   * @param string fsearch_type "category", "parameter"
   * @param int depth how far down in a hierarchy to traverse from "startclass"
   * 
   * 
   * @todo tests and handling for large datasets for categories, classes
   * @todo more complex category arrangment based on RDF graphs rather than a 
   *   faceted search hiearchy only
   *  
   * @see getCustomSearchParamsData()
   * */
  public function getSearchParamsData($start_class, $fsearch_type, $depth = 3) {
    //TODO check search type
    return $this->getCustomSearchParamsData($start_class, $fsearch_type, $depth);
  }

  /**
  * a hack until the taxonomy integration is completed
  * */
  function taxonomyForm($allowed_uris) {
    $final_values = array ();
    $values = $this->getCustomSearchParamsData('property', null, null, null);
    //dbg($values,'$values');
    foreach ($values as $property => $data) {
      if (in_array(trim($data['label']), $allowed_uris)) {
        $final_values[$property] = $data;

      } else {
        // dbg($data['label'],"\$data['label']  $property");
      }
    }
    return $final_values;
  }
  
  /**
   * returns  a hierarchy of classes for 2 levels under a starting hiararchy
   * */
  private function getTelescopingClasses($startclass){
    $original_startclass= null;
    //catch overriding start class here
    if (is_array($this->sonia_search_array) && !empty($this->sonia_search_array)){
      foreach($this->sonia_search_array as $search_statement){
        if (isset($search_statement['category_query']) && $search_statement['category_query']== 1){
          $original_startclass = $startclass;
          $startclass = SONIA_H :: decodeURI($search_statement['object']);
        }
      }
    }
    return $this->store->getTelescopingClassesAndCountsHierarchy( $startclass,$original_startclass, true);
    //return $classes;
  }
  
  /**
   * A custom paramter value or place in a hiearchy that remembers either it's 
   * depth in the hiearchy or range of values within the total search results. 
   * 
   * @param string $query_type 'class', 'property', 'object'
   * @param mixed $uris the set of uris
   * @param string $return_type 
   *   'list_objects' 'list_properties' 'list_values'
   * @param array $values['min'], $values['max'] range search 
   *   null for all $values['regex']
   * @param string $order_by eg. 'label ASC'
   * @param boolean $return_counts include count information, can be a performance hit
   * 
   * @param boolean cache whether or not to cache the return value.
   * @param int $break_custom the in the telescoping search, how many breaks between 
   *  MIN and MAX should be set. MIN and Max can be integer or alphabetical
   * 
   * @return array ''
   *  
   * @see getSearchParamsData()
   * */
  public function getCustomSearchParamsData($query_type, $uris, $return_type, $values, $order_by = null, $return_counts = true, $depth = 3, $break_custom = 5, $cache = true) {
    $return_array = array ();
    
    //what if a query involves custom classes or properties?
    //for now assuming queries and classes don't need both
    switch ($query_type) {
      case 'class' :
        if ($return_type == 'telescope'){
          //travers search params to get current start class
          $results = $this->getTelescopingClasses($uris[0]);
          //dbg($results,'tele');
        }
        else{
          $results = $this->classQuery(null);
        }

        break;
      case 'property' :
        if (empty ($this->t_all_properties)) {
          $results = $this->propertyQuery();
          $this->t_all_properties = $results;
        } else {
          $results = $this->t_all_properties;
        }
        break;
      case 'p_value' :
        $results = $this->propertyValuesQuery($uris);
        //dbg($results,'pval');
        break;
      case 'property_classes' :
        //$this->sonia->store->getAllClassesOfProperty($property1);
        break;
      case 'object' :
        // classes and values?
        break;
      default :
        $message = $query_type . ' is not a valid $query_type for getCustomSearchParamsData()';
        $this->io->throwError($message);
        break;
    }

    //temp
    return $results;

  }

  //public function 

  /**
   * sets some variables for IO
   * */
  private function _setIORoutine() {
    global $use_search_cache;
    if (!isset ($use_search_cache))
      $use_search_cache = true;

    $this->io->setSearchParams($this->request_param, $use_search_cache);

    $this->search_params = $this->io->getSearchParams();
    $this->all_params = $this->io->getRequestValues();
  }

  //--------------------------------------------------------------------------
  /**
   * 
   * sets internal member variables and other standard initializations
   * must be called in the subclass via the contstructor
   * 
   * @return array a list of available implemented stores
   * 
   * */
  //--------------------------------------------------------------------------
  private function _init($store, $io) {
    if (SONIA_RUN_INIT_TESTS)
      $this->runTests();
    //create the store object
    $this->_setStore($store, '');
    //create the IO object
    $this->_setIO($io);
  }

  //--------------------------------------------------------------------------
  /**
  * sets the IO object factory pattern
  * */
  //--------------------------------------------------------------------------
  private function _setIO($io) {
    $this->io = new SONIA_IO();

    //TEST
    if (!($this->io instanceof SONIA_IO_API)) {
      trigger_error('IO object is wrong kind of object');
    }

    //HACK
    $this->io->setRequest(null);
  }

  //--------------------------------------------------------------------------
  /**
   * 
   * choose the RDF, DB, or triple backend implementation
   * a factory method returning a Sonia Object of the desired implementation
   * 
   * The param is a string aliases in the config file
   * 
   * Examples of RDF stores available
   *  - SONIA MySQL local
   *  - ARC - MySQL local
   *  - ARC - remote
   *  - other RDF STORE joseki remote
   * 
   * USAGE:
   * 
   * @todo should throw errors
   * @param string $store_type of the store - 
   * should be in the config in the following format 
   * 
   * @return void
   * 
   * */
  //--------------------------------------------------------------------------
  private function _setStore($store_type, $store_info) {
    switch ($store_type){
    	case 'sonia_sesame':
    	  $this->store = new SONIASesameStore($store_info);
    	  break;
    	case 'sonia_arc_local':
    	default:
		  	$this->store = new SONIAARCStore($store_info);
		  	break;
    
    }
 
    if (!($this->store instanceof SONIAStoreAPI)) {
      $this->io->throwError('store is wrong kind of object');
    }
  }

  //SEARCH

  /**
  * converts incoming RequestURI to SPARQL
  * @todo convert to callback
  * */
  public function parseRequest() {
//    
//     global $exectution_order;
//  
//     $exectution_order++;
//     dbg($exectution_order,'parseRequest $exectution_order');
  
    $this->sonia_search_array = array ();

    if (isset ($this->search_params[SONIA_FORM_REQUEST_PREFIX]))
      $this->search_params = $this->search_params[SONIA_FORM_REQUEST_PREFIX];

    if (is_array($this->search_params)) {
      foreach ($this->search_params as $key => $value) {
        if ($value !== '') {
          //switch ($key) {
            //default :
              if (is_array($value) && !empty ($value)) {
                if ($key == 'cat') {
                  foreach ($value as $k => $v) {
                   
                    if (!is_array($v)){
                      if ($v != 0){
                        $v = SONIA_H :: decodeURI($v);
                        $this->sonia_search_array[] = array (
                          'subject' => null,
                          'predicate' => SONIA_RDF_CREATE_TYPE,
                          'object' => $v,
                          'literal_object' => false,
                          'category_query' => true
                        );
                      }
                    } else   {
                      foreach($v as $vv){
                        
                      if (!empty($vv)){
                          $vv = SONIA_H :: decodeURI($vv);
                          $this->sonia_search_array[] = array (
                            'subject' => null,
                            'predicate' => SONIA_RDF_CREATE_TYPE,
                            'object' => $vv,
                            'literal_object' => false,
                            'category_query' => true
                          );
                        }
                      }
                    }
                    //}
                  }
                } else
                  if ($key == 'facet') {
                   
                    foreach ($value as $k => $v) {
                      if (!is_array($v)){
                          $v = SONIA_H :: decodeURI($v);
                          $this->sonia_search_array[] = array (
                            'subject' => null,
                            'predicate' => $v,
                            'object' => null,
                            'literal_object' => false
                          );
                        } else {
                           foreach($v as $vv){
                              if (!empty($vv)){
                                  $vv = SONIA_H :: decodeURI($vv);
                                  $this->sonia_search_array[] = array (
                                    'subject' => null,
                                    'predicate' => $vv,
                                    'object' => null,
                                    'literal_object' => false 
                                  );
                                }
                              }
                      }
                    } 
                  } else
                    if ($key == 'p_value') {
                      foreach ($value as $k => $v) {
                        if (!is_array($v)) {
                          $k = SONIA_H :: decodeURI($k);
                          $v = SONIA_H :: decodeURI($v);
                          if ((!empty ($k)) && ($v != '') && ($v != '0')) {
                            if (strstr($v, SONIA_DEFAULT_MODEL_IRI) || strstr($v, SONIA_RDF_URI) || strstr($v, SONIA_RDFS_URI)) {
                              $this->sonia_search_array[] = array (
                                'subject' => null,
                                'predicate' => $v,
                                'object' => $k,
                                'literal_object' => true,
                                'literal_type' => 'exact'
                              );
                            } else {
                              $this->sonia_search_array[] = array (
                                'subject' => null,
                                'predicate' => $k,
                                'object' => $v,
                                'literal_object' => true,
                                'literal_type' => 'exact'
                              );
                            }
                          }
                        } else { //allow multiple
                          foreach ($v as $kk => $vv) {
                            $kk = SONIA_H :: decodeURI($kk);
                            $vv = SONIA_H :: decodeURI($vv);
                            if ((!empty ($kk)) && ($vv != '') && ($vv != '0')) {
                              if (strstr($vv, SONIA_DEFAULT_MODEL_IRI) || strstr($vv, SONIA_RDF_URI) || strstr($vv, SONIA_RDFS_URI)) {
                                $this->sonia_search_array[] = array (
                                  'subject' => null,
                                  'predicate' => $vv,
                                  'object' => $kk,
                                  'literal_object' => true,
                                  'literal_type' => 'exact'
                                );
                              } else {
                                $this->sonia_search_array[] = array (
                                  'subject' => null,
                                  'predicate' => $kk,
                                  'object' => $vv,
                                  'literal_object' => true,
                                  'literal_type' => 'exact'
                                );
                              }
                            }
                          }
                        }

                      }
                    } else
                      if ('qstring') {
                        foreach ($value as $k => $v) {
                          if (!empty ($v) && !is_array($v)) {
                            $predicate = (is_int($k)) ? null : $k;
                            $this->sonia_search_array[] = array (
                              'subject' => null,
                              'predicate' => $predicate,
                              'object' => $v,
                              'literal_object' => true,
                              'literal_type' => 'fulltext'
                            );
                          } else {
                            foreach ($v as $kk => $vv) {
                              $predicate = (is_int($kk)) ? null : $kk;
                              if (!empty ($vv)) {
                                $this->sonia_search_array[] = array (
                                  'subject' => null,
                                  'predicate' => $predicate,
                                  'object' => $vv,
                                  'literal_object' => true,
                                  'literal_type' => 'fulltext'
                                );
                              }
                            }
                          }

                        }
                      } else {

                      }
              }

             //dbg($this->sonia_search_array, '$this->sonia_search_array');

              /*
              if (isset($this->sonia->all_categories[$key])) 
              {
                 
                 $predicate = str_replace(' ','_',$this->sonia->all_categories[$key]['name']);
                 $this->sonia_search_array[] = array ('subject'=>null,'predicate'=> $predicate,'object'=>null,'literal_object'=>false); 
              }
              */

            //  break;
          //} //end switch

        } //end if

      } //end foreach
    } //end if
  }

  //---------------------------------------------------------------------------
  /**
   * Returns the results of a search
   * */
  //---------------------------------------------------------------------------
  public function getSearchResults() {
    return isset ($this->search_results) ? $this->search_results : null;
  }


  //--------------------------------------------------------------------------
  /**
   * does some basic integrety checks. Is meant to be run during regular
  * program execution, but will slow down the app a bit.
  * 
   * run if the config option SONIA_RUN_INIT_TESTS is enabled in the 
   * implemntation.
  * 
  * For running complete tests see the /tests directory 
   * 
   * @return array a list of available implemented stores
   * 
   * */
  //--------------------------------------------------------------------------
  private function runTests() {
    //values of defines that should not be re-used in new statements
    $this->reserved_predicates = array (
      SONIA_TYPE,
      SONIA_NAME,
      SONIA_CONTAINER,
      SONIA_CLASS,
      SONIA_PROPERTY,
      SONIA_RDF_CREATE_TYPE,
      SONIA_SUBCLASS,
      SONIA_DOMAIN,
      SONIA_DEFAULT_CATEGORY_START,
      SONIA_NEEDS,
      APP_DIRECTORY,

      
    );

    //defines that should be set and will be tested
    $this->reserved_defines = array (
      'SONIA_TYPE',
      'SONIA_NAME',
      'SONIA_CONTAINER',
      'SONIA_CLASS',
      'SONIA_PROPERTY',
      'SONIA_RDF_CREATE_TYPE',
      'SONIA_SUBCLASS',
      'SONIA_DOMAIN',
      'SONIA_DEFAULT_CATEGORY_START',
      'SONIA_NEEDS',
      'APP_DIRECTORY',

      
    );

    //check defines
    foreach ($this->reserved_defines as $define) {
      if (!defined($define)) {
        $this->io->throwError($define . ' not defined.');
      }
    }

  }

  //--------------------------------------------------------------------------
  /**
   * 
   * provides a list of back stores available
   * @return array a list of available implemented stores
   * 
   * */
  //--------------------------------------------------------------------------
  public function listStores() {
    //not implemented
  }

  //SEARCH
  function getDebugText() {
    return $this->store->debug_text;
  }

  //--------------------------------------------------------------------------
  /**
   * Export data to a file and a native format
   * 
   * @param array query given a query, export the results to a
   * @return string flat file conents
   * 
   * */
  //--------------------------------------------------------------------------
  public function exportNative() {
    //get results of query either that is based on search or RDF
    if (!empty ($this->sonia_search_array) && is_array(empty ($this->sonia_search_array))) {
      $results = $this->query(0, $this->sonia_search_array, null, null, true, '', '');
    } else {
      $results = $this->store->exportNative();
    }
    //parse file format
    //return string
    return $this->parseNativeTripleToString($results);
  }

  //--------------------------------------------------------------------------
  /**
   * Import data from a SONIA native triple format given one of four methods.
   * 
   * @param string method 1. overwrite old - no duplicates 2. preserve old - not duplicates 3. duplicates allowed 4. clear out ALL old
   * @param string file, filecontents uplodaded
   * @return array verbose output string succes or failure
   * */
  //--------------------------------------------------------------------------
  public function importNative($method, $file) {

  }

  //--------------------------------------------------------------------------
  /**
  * import RDF into the rdf store 
  * 
  * @todo validate the RDF
  * @param string $iri_model the name of the model
  * @param string $file the location of the file, can be any uri PHP allows http:// file:/// or just a path
  * */
  //--------------------------------------------------------------------------
  public function importRDF($iri_model, $file) {
    return $this->store->importRDF($iri_model, $file);
  }

  //exportRDF($iri_model)

  //--------------------------------------------------------------------------
  /**
  * export the rdf store 
  * 
  * @todo validate the RDF
  * @param string $iri_model the name of the model
  * */
  //--------------------------------------------------------------------------
  public function exportRDF($iri_model) {
    return $this->store->exportRDF($iri_model);
  }

}
